from __future__ import annotations

from enum import Enum
from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Type, Union

import farmhash
import numpy as np
from arango.aql import Cursor
from arango.database import StandardDatabase
from arango.exceptions import ArangoServerError, ViewGetError
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from langchain_core.vectorstores.utils import maximal_marginal_relevance
from packaging import version

from langchain_arangodb.vectorstores.utils import DistanceStrategy

DEFAULT_DISTANCE_STRATEGY = DistanceStrategy.COSINE
DISTANCE_MAPPING = {
    DistanceStrategy.EUCLIDEAN_DISTANCE: "l2",
    DistanceStrategy.COSINE: "cosine",
}


class SearchType(str, Enum):
    """Enumerator of the search types."""

    VECTOR = "vector"
    HYBRID = "hybrid"


DEFAULT_SEARCH_TYPE = SearchType.VECTOR

# Constants for RRF
DEFAULT_RRF_CONSTANT = 60  # Standard constant for RRF
DEFAULT_SEARCH_LIMIT = 100  # Default limit for initial search results

# Full-text search analyzer options
DEFAULT_ANALYZER = "text_en"  # Default analyzer for full-text search
SUPPORTED_ANALYZERS = [
    "text_en",
    "text_de",
    "text_es",
    "text_fi",
    "text_fr",
    "text_it",
    "text_nl",
    "text_no",
    "text_pt",
    "text_ru",
    "text_sv",
    "text_zh",
]


class ArangoVector(VectorStore):
    """ArangoDB vector store implementation for LangChain.

    This class provides a vector store implementation using ArangoDB as the backend.
    It supports both vector similarity search and hybrid search (vector + keyword)
    capabilities.

    Args:
        embedding: The embedding function to use for converting text to vectors.
            Must implement the `langchain.embeddings.base.Embeddings` interface.
        embedding_dimension: The dimensionality of the embedding vectors.
            Must match the output dimension of the embedding function.
        database: The ArangoDB database instance to use for storage and retrieval.
        collection_name: The name of the ArangoDB collection to store
            documents. Defaults to "documents".
        search_type: The type of search to perform. Can be either "vector" for pure
            vector similarity search or "hybrid" for combining vector and
            keyword search. Defaults to "vector".
        embedding_field: The field name in the document to store the embedding vector.
            Defaults to "embedding".
        text_field: The field name in the document to store the text content.
            Defaults to "text".
        vector_index_name: The name of the vector index to create in ArangoDB.
            This index enables efficient vector similarity search.
            Defaults to "vector_index".
        distance_strategy: The distance metric to use for vector similarity.
            Can be either "COSINE" or "EUCLIDEAN_DISTANCE".
            Defaults to "COSINE".
        num_centroids: The number of centroids to use for the vector index.
            Higher values can improve search accuracy but increase memory usage.
            Defaults to 1.
        relevance_score_fn: Optional function to normalize the relevance score.
            If not provided, uses the default normalization for the distance strategy.
        keyword_index_name: The name of the ArangoDB View created to enable
            Full-Text-Search capabilities. Only used if search_type is set
            to "hybrid". Defaults to "keyword_index".
        keyword_analyzer: The text analyzer to use for keyword search.
            Must be one of the supported analyzers in ArangoDB.
            Defaults to "text_en".
        rrf_constant: The constant used in Reciprocal Rank Fusion (RRF) for hybrid
            search. Higher values give more weight to lower-ranked results.
            Defaults to 60.
        rrf_search_limit: The maximum number of results to consider in RRF scoring.
            Defaults to 100.

    Example:
        .. code-block:: python

            from arango import ArangoClient
            from langchain_community.embeddings.openai import OpenAIEmbeddings
            from langchain_community.vectorstores.arangodb_vector import ArangoVector

            # Initialize ArangoDB connection
            db = ArangoClient("http://localhost:8529").db(
                "test",
                username="root",
                password="openSesame"
            )

            # Create embedding function
            embedding = OpenAIEmbeddings(
                model="text-embedding-3-small",
                dimensions=dimension
            )

            # Create vector store
            vector_store = ArangoVector.from_texts(
                texts=["hello world", "hello langchain", "hello arangodb"],
                embedding=embedding,
                database=db,
                collection_name="Documents"
            )

            # Perform similarity search
            print(vector_store.similarity_search("arangodb", k=1))
    """

    def __init__(
        self,
        embedding: Embeddings,
        embedding_dimension: int,
        database: StandardDatabase,
        collection_name: str = "documents",
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        embedding_field: str = "embedding",
        text_field: str = "text",
        vector_index_name: str = "vector_index",
        distance_strategy: DistanceStrategy = DEFAULT_DISTANCE_STRATEGY,
        num_centroids: int = 1,
        relevance_score_fn: Optional[Callable[[float], float]] = None,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
    ):
        if search_type not in [SearchType.VECTOR, SearchType.HYBRID]:
            raise ValueError("search_type must be 'vector' or 'hybrid'")

        if distance_strategy not in [
            DistanceStrategy.COSINE,
            DistanceStrategy.EUCLIDEAN_DISTANCE,
        ]:
            m = "distance_strategy must be 'COSINE' or 'EUCLIDEAN_DISTANCE'"
            raise ValueError(m)

        self.embedding = embedding
        self.embedding_dimension = int(embedding_dimension)
        self.db = database
        self.async_db = self.db.begin_async_execution(return_result=False)
        self.search_type = search_type
        self.collection_name = collection_name
        self.embedding_field = embedding_field
        self.text_field = text_field
        self.vector_index_name = vector_index_name
        self._distance_strategy = distance_strategy
        self.num_centroids = num_centroids
        self.override_relevance_score_fn = relevance_score_fn

        # Hybrid search parameters
        self.keyword_index_name = keyword_index_name
        self.keyword_analyzer = keyword_analyzer
        self.rrf_constant = rrf_constant
        self.rrf_search_limit = rrf_search_limit

        if not self.db.has_collection(collection_name):
            self.db.create_collection(collection_name)

        self.collection = self.db.collection(self.collection_name)

    @property
    def embeddings(self) -> Embeddings:
        return self.embedding

    def retrieve_vector_index(self) -> Union[dict[str, Any], None]:
        """Retrieve the vector index from the collection."""
        indexes = self.collection.indexes()  # type: ignore
        for index in indexes:  # type: ignore
            if index["name"] == self.vector_index_name:
                return index

        return None

    def create_vector_index(self) -> None:
        """Create the vector index on the collection."""
        self.collection.add_index(  # type: ignore
            {
                "name": self.vector_index_name,
                "type": "vector",
                "fields": [self.embedding_field],
                "params": {
                    "metric": DISTANCE_MAPPING[self._distance_strategy],
                    "dimension": self.embedding_dimension,
                    "nLists": self.num_centroids,
                },
            }
        )

    def delete_vector_index(self) -> None:
        """Delete the vector index from the collection."""
        index = self.retrieve_vector_index()

        if index is not None:
            self.collection.delete_index(index["id"])

    def retrieve_keyword_index(self) -> Optional[dict[str, Any]]:
        """Retrieve the keyword index from the collection."""
        try:
            return self.db.view(self.keyword_index_name)  # type: ignore
        except ViewGetError:
            return None

    def create_keyword_index(self) -> None:
        """Create the keyword index on the collection."""
        if self.retrieve_keyword_index():
            return

        view_properties = {
            "links": {
                self.collection_name: {
                    "analyzers": [self.keyword_analyzer],
                    "fields": {self.text_field: {"analyzers": [self.keyword_analyzer]}},
                }
            }
        }

        self.db.create_view(self.keyword_index_name, "arangosearch", view_properties)

    def delete_keyword_index(self) -> None:
        """Delete the keyword index from the collection."""
        view = self.retrieve_keyword_index()
        if view:
            self.db.delete_view(self.keyword_index_name)

    def add_embeddings(
        self,
        texts: Iterable[str],
        embeddings: List[List[float]],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        batch_size: int = 500,
        use_async_db: bool = False,
        insert_text: bool = True,
        **kwargs: Any,
    ) -> List[str]:
        """Add embeddings to the vectorstore."""
        texts = list(texts)

        if ids is None:
            ids = [str(farmhash.Fingerprint64(text.encode("utf-8"))) for text in texts]  # type: ignore

        if not metadatas:
            metadatas = [{} for _ in texts]

        if len(ids) != len(texts) != len(embeddings) != len(metadatas):
            m = "Length of ids, texts, embeddings and metadatas must be the same."
            raise ValueError(m)

        db = self.async_db if use_async_db else self.db
        collection = db.collection(self.collection_name)

        data = []
        for _key, text, embedding, metadata in zip(ids, texts, embeddings, metadatas):
            doc: dict[str, Any] = {self.text_field: text} if insert_text else {}

            doc.update(
                {
                    **metadata,
                    "_key": _key,
                    self.embedding_field: embedding,
                }
            )

            data.append(doc)

            if len(data) == batch_size:
                collection.import_bulk(data, on_duplicate="update", **kwargs)
                data = []

        collection.import_bulk(data, on_duplicate="update", **kwargs)

        return ids

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Add texts to the vector store.

        This method embeds the provided texts using the embedding function and stores
        them in ArangoDB along with their embeddings and metadata.

        Args:
            texts: An iterable of text strings to add to the vector store.
            metadatas: Optional list of metadata dictionaries to associate with each
                text. Each dictionary can contain arbitrary key-value pairs that
                will be stored alongside the text and embedding.
            ids: Optional list of unique identifiers for each text. If not provided,
                IDs will be generated using a hash of the text content.
            **kwargs: Additional keyword arguments passed to add_embeddings.

        Returns:
            List of document IDs that were added to the vector store.
        """
        embeddings = self.embedding.embed_documents(list(texts))

        return self.add_embeddings(
            texts=texts, embeddings=embeddings, metadatas=metadatas, ids=ids, **kwargs
        )

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        filter_clause: str = "",
        search_type: Optional[SearchType] = None,
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
        **kwargs: Any,
    ) -> List[Document]:
        """Search for similar documents using vector similarity or hybrid search.

        This method performs a similarity search using either pure vector similarity
        or a hybrid approach combining vector and keyword search. The search type
        can be overridden for individual queries.

        Args:
            query: The text query to search for.
            k: The number of most similar documents to return. Defaults to 4.
            return_fields: Set of additional document fields to return in results.
                The _key and text fields are always returned.
            use_approx: Whether to use approximate nearest neighbor search.
                Enables faster but potentially less accurate results.
                Defaults to True.
            embedding: Optional pre-computed embedding for the query.
                If not provided, the query will be embedded using the embedding
                function.
            filter_clause: Optional AQL filter clause to apply to the search.
                Can be used to filter results based on document properties.
            search_type: Override the default search type for this query.
                Can be either "vector" or "hybrid".
            vector_weight: Weight to apply to vector similarity scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_weight: Weight to apply to keyword search scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_search_clause: Optional AQL filter clause to apply Full Text Search.
                If empty, a default search clause will be used.

        Returns:
            List of Document objects most similar to the query.
        """
        search_type = search_type or self.search_type
        embedding = embedding or self.embedding.embed_query(query)

        if search_type == SearchType.VECTOR:
            return self.similarity_search_by_vector(
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
            )

        else:
            return self.similarity_search_by_vector_and_keyword(
                query=query,
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                vector_weight=vector_weight,
                keyword_weight=keyword_weight,
                keyword_search_clause=keyword_search_clause,
            )

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        filter_clause: str = "",
        search_type: Optional[SearchType] = None,
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Search for similar documents and return their similarity scores.

        Similar to similarity_search but returns a tuple of (Document, score) for each
        result. The score represents the similarity between the query and the document.

        Args:
            query: The text query to search for.
            k: The number of most similar documents to return. Defaults to 4.
            return_fields: Set of additional document fields to return in results.
                The _key and text fields are always returned.
            use_approx: Whether to use approximate nearest neighbor search.
                Enables faster but potentially less accurate results.
                Defaults to True.
            embedding: Optional pre-computed embedding for the query.
                If not provided, the query will be embedded using the embedding
                function.
            filter_clause: Optional AQL filter clause to apply to the search.
                Can be used to filter results based on document properties.
            search_type: Override the default search type for this query.
                Can be either "vector" or "hybrid".
            vector_weight: Weight to apply to vector similarity scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_weight: Weight to apply to keyword search scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_search_clause: Optional AQL filter clause to apply Full Text Search.
                If empty, a default search clause will be used.

        Returns:
            List of tuples containing (Document, score) pairs, sorted by score.
        """
        search_type = search_type or self.search_type
        embedding = embedding or self.embedding.embed_query(query)

        if search_type == SearchType.VECTOR:
            return self.similarity_search_by_vector_with_score(
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
            )

        else:
            return self.similarity_search_by_vector_and_keyword_with_score(
                query=query,
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                vector_weight=vector_weight,
                keyword_weight=keyword_weight,
                keyword_search_clause=keyword_search_clause,
            )

    def similarity_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        **kwargs: Any,
    ) -> List[Document]:
        """Return docs most similar to embedding vector.

        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            return_fields: Fields to return in the result. For example,
                {"foo", "bar"} will return the "foo" and "bar" fields of the document,
                in addition to the _key & text field. Defaults to an empty set.
            use_approx: Whether to use approximate vector search via ANN.
                Defaults to True. If False, exact vector search will be used.
            filter_clause: Filter clause to apply to the query.

        Returns:
            List of Documents most similar to the query vector.
        """
        results = self.similarity_search_by_vector_with_score(
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
        )

        return [doc for doc, _ in results]

    def similarity_search_by_vector_and_keyword(
        self,
        query: str,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
    ) -> List[Document]:
        results = self.similarity_search_by_vector_and_keyword_with_score(
            query=query,
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            vector_weight=vector_weight,
            keyword_weight=keyword_weight,
            keyword_search_clause=keyword_search_clause,
        )

        return [doc for doc, _ in results]

    def similarity_search_by_vector_with_score(
        self,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Return docs most similar to embedding vector.

        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            return_fields: Fields to return in the result. For example,
                {"foo", "bar"} will return the "foo" and "bar" fields of the document,
                in addition to the _key & text field. Defaults to an empty set.
            use_approx: Whether to use approximate vector search via ANN.
                Defaults to True. If False, exact vector search will be used.
            filter_clause: Filter clause to apply to the query.
            **kwargs: Additional keyword arguments passed to the query execution.

        Returns:
            List of Documents most similar to the query vector.
        """
        aql_query, bind_vars = self._build_vector_search_query(
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
        )

        cursor = self.db.aql.execute(aql_query, bind_vars=bind_vars, stream=True)

        results = self._process_search_query(cursor)  # type: ignore

        return results

    def similarity_search_by_vector_and_keyword_with_score(
        self,
        query: str,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Run similarity search with ArangoDB.

        Args:
            query (str): Query text to search for.
            k (int): Number of results to return. Defaults to 4.
            return_fields: Fields to return in the result. For example,
                {"foo", "bar"} will return the "foo" and "bar" fields of the document,
                in addition to the _key & text field. Defaults to an empty set.
            use_approx: Whether to use approximate vector search via ANN.
                Defaults to True. If False, exact vector search will be used.
            filter_clause: Filter clause to apply to the query.
            vector_weight: Weight to apply to vector similarity scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_weight: Weight to apply to keyword search scores in hybrid search.
                Only used when search_type is "hybrid". Defaults to 1.0.
            keyword_search_clause: Optional AQL filter clause to apply Full Text Search.
                If empty, a default search clause will be used.

        Returns:
            List of Documents most similar to the query.
        """

        aql_query, bind_vars = self._build_hybrid_search_query(
            query=query,
            k=k,
            embedding=embedding,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            vector_weight=vector_weight,
            keyword_weight=keyword_weight,
            keyword_search_clause=keyword_search_clause,
        )

        cursor = self.db.aql.execute(aql_query, bind_vars=bind_vars, stream=True)

        results = self._process_search_query(cursor)  # type: ignore

        return results

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        """Delete by vector ID or other criteria.

        Args:
            ids: List of ids to delete.
            **kwargs: Other keyword arguments that can be used to delete vectors.

        Returns:
            Optional[bool]: True if deletion is successful,
                None if no ids are provided, or raises an exception if an error occurs.
        """
        if not ids:
            return None

        for result in self.collection.delete_many(ids, **kwargs):  # type: ignore
            if isinstance(result, ArangoServerError):
                raise result

        return True

    def get_by_ids(self, ids: Sequence[str], /) -> list[Document]:
        """Get documents by their IDs.

        Args:
            ids: List of ids to get.

        Returns:
            List of Documents with the given ids.
        """
        docs = []
        doc: dict[str, Any]

        for doc in self.collection.get_many(ids):  # type: ignore
            _key = doc.pop("_key")
            page_content = doc.pop(self.text_field)

            docs.append(Document(page_content=page_content, id=_key, metadata=doc))

        return docs

    def max_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Search for documents using Maximal Marginal Relevance (MMR).

        MMR optimizes for both similarity to the query and diversity among the results.
        It helps avoid returning redundant or very similar documents.

        Args:
            query: The text query to search for.
            k: The number of documents to return. Defaults to 4.
            fetch_k: The number of documents to fetch for MMR selection.
                Should be larger than k to allow for diversity selection.
                Defaults to 20.
            lambda_mult: Controls the diversity vs relevance tradeoff.
                Values between 0 and 1, where:
                - 0: Maximum diversity
                - 1: Maximum relevance
                Defaults to 0.5.
            return_fields: Set of additional document fields to return in results.
                The _key and text fields are always returned.
            use_approx: Whether to use approximate nearest neighbor search.
                Enables faster but potentially less accurate results.
                Defaults to True.
            embedding: Optional pre-computed embedding for the query.
                If not provided, the query will be embedded using the embedding
                function.
            **kwargs: Additional keyword arguments passed to the search methods.

        Returns:
            List of Document objects selected by MMR algorithm.
        """
        return_fields.add(self.embedding_field)

        # Embed the query
        query_embedding = embedding or self.embedding.embed_query(query)

        # Fetch the initial documents
        docs = self.similarity_search_by_vector(
            embedding=query_embedding,
            k=fetch_k,
            return_fields=return_fields,
            use_approx=use_approx,
            **kwargs,
        )

        # Get the embeddings for the fetched documents
        embeddings = [doc.metadata[self.embedding_field] for doc in docs]

        # Select documents using maximal marginal relevance
        selected_indices = maximal_marginal_relevance(
            np.array(query_embedding), embeddings, lambda_mult=lambda_mult, k=k
        )

        selected_docs = [docs[i] for i in selected_indices]

        return selected_docs

    def _select_relevance_score_fn(self) -> Callable[[float], float]:
        if self.override_relevance_score_fn is not None:
            return self.override_relevance_score_fn

        # Default strategy is to rely on distance strategy provided
        # in vectorstore constructor
        if self._distance_strategy in [
            DistanceStrategy.COSINE,
            DistanceStrategy.EUCLIDEAN_DISTANCE,
        ]:
            return lambda x: x
        else:
            raise ValueError(
                "No supported normalization function"
                f" for distance_strategy of {self._distance_strategy}."
                "Consider providing relevance_score_fn to ArangoVector constructor."
            )

    @classmethod
    def from_texts(
        cls: Type[ArangoVector],
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        database: Optional[StandardDatabase] = None,
        collection_name: str = "documents",
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        embedding_field: str = "embedding",
        text_field: str = "text",
        index_name: str = "vector_index",
        distance_strategy: DistanceStrategy = DEFAULT_DISTANCE_STRATEGY,
        num_centroids: int = 1,
        ids: Optional[List[str]] = None,
        overwrite_index: bool = False,
        insert_text: bool = True,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
        **kwargs: Any,
    ) -> ArangoVector:
        """Create an ArangoVector instance from a list of texts.

        This is a convenience method that creates a new ArangoVector instance,
        embeds the provided texts, and stores them in ArangoDB.

        Args:
            texts: List of text strings to add to the vector store.
            embedding: The embedding function to use for converting text to vectors.
            metadatas: Optional list of metadata dictionaries to associate with each
                text.
            database: The ArangoDB database instance to use.
            collection_name: The name of the ArangoDB collection to use.
                Defaults to "documents".
            search_type: The type of search to perform. Can be either "vector" or
                "hybrid". Defaults to "vector".
            embedding_field: The field name to store embeddings. Defaults to
                "embedding".
            text_field: The field name to store text content. Defaults to "text".
            index_name: The name of the vector index. Defaults to "vector_index".
            distance_strategy: The distance metric to use. Defaults to "COSINE".
            num_centroids: Number of centroids for vector index. Defaults to 1.
            ids: Optional list of unique identifiers for each text.
            overwrite_index: Whether to delete and recreate existing indexes.
                Defaults to False.
            insert_text: Whether to store the text content in the database.
                Required for hybrid search. Defaults to True.
            keyword_index_name: Name of the keyword search index. Defaults to
                "keyword_index".
            keyword_analyzer: Text analyzer for keyword search. Defaults to "text_en".
            rrf_constant: Constant for RRF scoring in hybrid search. Defaults to 60.
            rrf_search_limit: Maximum results for RRF scoring. Defaults to 100.
            **kwargs: Additional keyword arguments passed to the constructor.

        Returns:
            A new ArangoVector instance with the texts embedded and stored.
        """
        if not database:
            raise ValueError("Database must be provided.")

        if not insert_text and search_type == SearchType.HYBRID:
            raise ValueError("insert_text must be True when search_type is HYBRID")

        embeddings = embedding.embed_documents(list(texts))

        embedding_dimension = len(embeddings[0])

        store = cls(
            embedding,
            embedding_dimension=embedding_dimension,
            database=database,
            collection_name=collection_name,
            search_type=search_type,
            embedding_field=embedding_field,
            text_field=text_field,
            vector_index_name=index_name,
            distance_strategy=distance_strategy,
            num_centroids=num_centroids,
            keyword_index_name=keyword_index_name,
            keyword_analyzer=keyword_analyzer,
            rrf_constant=rrf_constant,
            rrf_search_limit=rrf_search_limit,
            **kwargs,
        )

        if overwrite_index:
            store.delete_vector_index()

            if search_type == SearchType.HYBRID:
                store.delete_keyword_index()

        store.add_embeddings(
            texts, embeddings, metadatas=metadatas, ids=ids, insert_text=insert_text
        )

        return store

    @classmethod
    def from_existing_collection(
        cls: Type[ArangoVector],
        collection_name: str,
        text_properties_to_embed: List[str],
        embedding: Embeddings,
        database: StandardDatabase,
        embedding_field: str = "embedding",
        text_field: str = "text",
        batch_size: int = 1000,
        aql_return_text_query: str = "",
        insert_text: bool = False,
        skip_existing_embeddings: bool = False,
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
        **kwargs: Any,
    ) -> ArangoVector:
        """Create an ArangoVector instance from an existing ArangoDB collection.

        This method reads documents from an existing collection, extracts specified
        text properties, embeds them, and creates a new vector store.

        Args:
            collection_name: Name of the existing ArangoDB collection.
            text_properties_to_embed: List of document properties containing text to
                embed. These properties will be concatenated to create the
                text for embedding.
            embedding: The embedding function to use for converting text to vectors.
            database: The ArangoDB database instance to use.
            embedding_field: The field name to store embeddings. Defaults to
                "embedding".
            text_field: The field name to store text content. Defaults to "text".
            batch_size: Number of documents to process in each batch. Defaults to 1000.
            aql_return_text_query: Custom AQL query to extract text from properties.
                Defaults to "RETURN doc[p]".
            insert_text: Whether to store the concatenated text in the database.
                Required for hybrid search. Defaults to False.
            skip_existing_embeddings: Whether to skip documents that already have
                embeddings. Defaults to False.
            search_type: The type of search to perform. Can be either "vector" or
                "hybrid". Defaults to "vector".
            keyword_index_name: Name of the keyword search index. Defaults to
                "keyword_index".
            keyword_analyzer: Text analyzer for keyword search. Defaults to "text_en".
            rrf_constant: Constant for RRF scoring in hybrid search. Defaults to 60.
            rrf_search_limit: Maximum results for RRF scoring. Defaults to 100.
            **kwargs: Additional keyword arguments passed to the constructor.

        Returns:
            A new ArangoVector instance with embeddings created from the collection.
        """
        if not text_properties_to_embed:
            m = "Parameter `text_properties_to_embed` must not be an empty list"
            raise ValueError(m)

        if text_field in text_properties_to_embed:
            m = "Parameter `text_field` must not be in `text_properties_to_embed`"
            raise ValueError(m)

        if not insert_text and search_type == SearchType.HYBRID:
            raise ValueError("insert_text must be True when search_type is HYBRID")

        if not aql_return_text_query:
            aql_return_text_query = "RETURN doc[p]"

        filter_clause = ""
        if skip_existing_embeddings:
            filter_clause = f"FILTER doc.{embedding_field} == null"

        aql_query = f"""
            FOR doc IN @@collection
                {filter_clause}

                LET texts = (
                    FOR p IN @properties
                        FILTER doc[p] != null
                        {aql_return_text_query}
                )

                RETURN {{
                    key: doc._key,
                    text: CONCAT_SEPARATOR(" ", texts),
                }}
        """

        bind_vars = {
            "@collection": collection_name,
            "properties": text_properties_to_embed,
        }

        cursor: Cursor = database.aql.execute(
            aql_query,
            bind_vars=bind_vars,  # type: ignore
            batch_size=batch_size,
            stream=True,
        )

        store: ArangoVector | None = None

        while not cursor.empty():
            batch = cursor.batch()
            batch_list = list(batch)  # type: ignore

            texts = [doc["text"] for doc in batch_list]
            ids = [doc["key"] for doc in batch_list]

            store = cls.from_texts(
                texts=texts,
                embedding=embedding,
                database=database,
                collection_name=collection_name,
                embedding_field=embedding_field,
                text_field=text_field,
                ids=ids,
                insert_text=insert_text,
                search_type=search_type,
                keyword_index_name=keyword_index_name,
                keyword_analyzer=keyword_analyzer,
                rrf_constant=rrf_constant,
                rrf_search_limit=rrf_search_limit,
                **kwargs,
            )

            batch.clear()  # type: ignore

            if cursor.has_more():
                cursor.fetch()

        if store is None:
            raise ValueError(f"No documents found in collection in {collection_name}")

        return store

    def _process_search_query(self, cursor: Cursor) -> List[tuple[Document, float]]:
        data: dict[str, Any]
        score: float
        results = []

        while not cursor.empty():
            for result in cursor:
                data, score = result["data"], result["score"]
                _key = data.pop("_key")
                page_content = data.pop(self.text_field)
                doc = Document(page_content=page_content, id=_key, metadata=data)

                results.append((doc, score))

            if cursor.has_more():
                cursor.fetch()

        return results

    def _build_vector_search_query(
        self,
        embedding: List[float],
        k: int,
        return_fields: set[str],
        use_approx: bool,
        filter_clause: str,
    ) -> Tuple[str, dict[str, Any]]:
        if self._distance_strategy == DistanceStrategy.COSINE:
            score_func = "APPROX_NEAR_COSINE" if use_approx else "COSINE_SIMILARITY"
            sort_order = "DESC"
        elif self._distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            score_func = "APPROX_NEAR_L2" if use_approx else "L2_DISTANCE"
            sort_order = "ASC"
        else:
            raise ValueError(f"Unsupported metric: {self._distance_strategy}")

        if use_approx:
            if version.parse(self.db.version()) < version.parse("3.12.4"):  # type: ignore
                m = "Approximate Nearest Neighbor search requires ArangoDB >= 3.12.4."
                raise ValueError(m)

            if not self.retrieve_vector_index():
                self.create_vector_index()

        return_fields.update({"_key", self.text_field})
        return_fields_list = list(return_fields)

        aql_query = f"""
            FOR doc IN @@collection
                {filter_clause if not use_approx else ""}
                LET score = {score_func}(doc.{self.embedding_field}, @embedding)
                SORT score {sort_order}
                LIMIT {k}
                {filter_clause if use_approx else ""}
                LET data = KEEP(doc, {return_fields_list})
                RETURN {{data, score}}
        """

        bind_vars = {
            "@collection": self.collection_name,
            "embedding": embedding,
        }

        return aql_query, bind_vars

    def _build_hybrid_search_query(
        self,
        query: str,
        k: int,
        embedding: List[float],
        return_fields: set[str],
        use_approx: bool,
        filter_clause: str,
        vector_weight: float,
        keyword_weight: float,
        keyword_search_clause: str,
    ) -> Tuple[str, dict[str, Any]]:
        """Build the hybrid search query using RRF."""

        if not self.retrieve_keyword_index():
            self.create_keyword_index()

        if self._distance_strategy == DistanceStrategy.COSINE:
            score_func = "APPROX_NEAR_COSINE" if use_approx else "COSINE_SIMILARITY"
            sort_order = "DESC"
        elif self._distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            score_func = "APPROX_NEAR_L2" if use_approx else "L2_DISTANCE"
            sort_order = "ASC"
        else:
            raise ValueError(f"Unsupported metric: {self._distance_strategy}")

        if use_approx:
            if version.parse(self.db.version()) < version.parse("3.12.4"):  # type: ignore
                m = "Approximate Nearest Neighbor search requires ArangoDB >= 3.12.4."
                raise ValueError(m)

            if not self.retrieve_vector_index():
                self.create_vector_index()

        return_fields.update({"_key", self.text_field})
        return_fields_list = list(return_fields)

        if not keyword_search_clause:
            keyword_search_clause = f"""
                SEARCH ANALYZER(
                    doc.{self.text_field} IN TOKENS(@query, @analyzer),
                    @analyzer
                )
            """

        aql_query = f"""
            LET vector_results = (
                FOR doc IN @@collection
                    {filter_clause if not use_approx else ""}
                    LET score = {score_func}(doc.{self.embedding_field}, @embedding)
                    SORT score {sort_order}
                    LIMIT {k}
                    {filter_clause if use_approx else ""}
                    RETURN {{ doc, score }}
            )

            LET keyword_results = (
                FOR doc IN @@view
                    {keyword_search_clause}
                    {filter_clause}
                    LET score = BM25(doc)
                    SORT score DESC
                    LIMIT {k}
                    RETURN {{ doc, score }}
            )

            LET rrf_vector = (
                FOR i IN RANGE(0, LENGTH(vector_results) - 1)
                    LET doc = vector_results[i].doc
                    FILTER doc != null
                    RETURN {{
                        doc,
                        score: {vector_weight} / (@rrf_constant + i + 1)
                    }}
            )

            LET rrf_keyword = (
                FOR i IN RANGE(0, LENGTH(keyword_results) - 1)
                    LET doc = keyword_results[i].doc
                    FILTER doc != null
                    RETURN {{
                        doc,
                        score: {keyword_weight} / (@rrf_constant + i + 1)
                    }}
            )

            FOR result IN APPEND(rrf_vector, rrf_keyword)
                COLLECT doc_key = result.doc._key INTO group
                LET rrf_score = SUM(group[*].result.score)
                LET doc = group[0].result.doc
                SORT rrf_score DESC
                LIMIT @rrf_search_limit
                RETURN {{
                    data: KEEP(doc, {return_fields_list}),
                    score: rrf_score
                }}
        """

        bind_vars = {
            "@collection": self.collection_name,
            "@view": self.keyword_index_name,
            "embedding": embedding,
            "query": query,
            "analyzer": self.keyword_analyzer,
            "rrf_constant": self.rrf_constant,
            "rrf_search_limit": self.rrf_search_limit,
        }

        return aql_query, bind_vars



import json
import os
import re
from collections import defaultdict
from math import ceil
from typing import Any, DefaultDict, Dict, List, Optional, Set, Union

import farmhash
import yaml
from arango import ArangoClient
from arango.database import Database, StandardDatabase
from arango.graph import Graph
from langchain_core.embeddings import Embeddings

from langchain_arangodb.graphs.graph_document import (
    Document,
    GraphDocument,
    Node,
    Relationship,
)
from langchain_arangodb.graphs.graph_store import GraphStore


def get_arangodb_client(
    url: Optional[str] = None,
    dbname: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Any:
    """Get the Arango DB client from credentials.

    Args:
        url: Arango DB url. Can be passed in as named arg or set as environment
            var ``ARANGODB_URL``. Defaults to "http://localhost:8529".
        dbname: Arango DB name. Can be passed in as named arg or set as
            environment var ``ARANGODB_DBNAME``. Defaults to "_system".
        username: Can be passed in as named arg or set as environment var
            ``ARANGODB_USERNAME``. Defaults to "root".
        password: Can be passed ni as named arg or set as environment var
            ``ARANGODB_PASSWORD``. Defaults to "".

    Returns:
        An arango.database.StandardDatabase.
    """
    _url: str = url or os.environ.get("ARANGODB_URL", "http://localhost:8529")  # type: ignore[assignment]
    _dbname: str = dbname or os.environ.get("ARANGODB_DBNAME", "_system")  # type: ignore[assignment]
    _username: str = username or os.environ.get("ARANGODB_USERNAME", "root")  # type: ignore[assignment]
    _password: str = password or os.environ.get("ARANGODB_PASSWORD", "")  # type: ignore[assignment]

    return ArangoClient(_url).db(_dbname, _username, _password, verify=True)


class ArangoGraph(GraphStore):
    """ArangoDB wrapper for graph operations.

    Parameters:
    - db (arango.database.StandardDatabase): ArangoDB database instance.
    - generate_schema_on_init (bool): Whether to generate the graph schema
        on initialization. Defaults to True.
    - schema_sample_ratio (float): A float (0 to 1) to determine the
        ratio of documents/edges sampled in relation to the Collection size
        to generate each Collection Schema. If 0, one document/edge
        is used per Collection. Defaults to 0.
    - schema_graph_name (str): The name of an existing ArangoDB Graph to specifically
        use to generate the schema. If None, the entire database will be used.
        Defaults to None.
    - schema_include_examples (bool): Whether to include example values fetched from
        a sample documents as part of the schema. Defaults to True. Lists of size
        higher than **schema_list_limit** will be excluded from the schema, even if
        **schema_include_examples** is set to True. Defaults to True.
    - schema_list_limit (int): The maximum list size the schema will include as part
        of the example values. If the list is longer than this limit, a string
        describing the list will be used in the schema instead. Default is 32.

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    def __init__(
        self,
        db: StandardDatabase,
        generate_schema_on_init: bool = True,
        schema_sample_ratio: float = 0,
        schema_graph_name: Optional[str] = None,
        schema_include_examples: bool = True,
        schema_list_limit: int = 32,
        schema_string_limit: int = 256,
    ) -> None:
        self.__db: StandardDatabase = db
        self.__async_db = db.begin_async_execution()

        self.__schema = {}
        if generate_schema_on_init:
            self.__schema = self.generate_schema(
                schema_sample_ratio,
                schema_graph_name,
                schema_include_examples,
                schema_list_limit,
                schema_string_limit,
            )

    @property
    def db(self) -> StandardDatabase:
        return self.__db

    @property
    def schema(self) -> Dict[str, Any]:
        """Returns the schema of the Graph Database as a structured object"""
        return self.__schema

    @property
    def get_structured_schema(self) -> Dict[str, Any]:
        """Returns the schema of the Graph Database as a structured object"""
        return self.__schema

    @property
    def schema_json(self) -> str:
        """Returns the schema of the Graph Database as a JSON string"""
        return json.dumps(self.__schema)

    @property
    def schema_yaml(self) -> str:
        """Returns the schema of the Graph Database as a YAML string"""
        return yaml.dump(self.__schema, sort_keys=False)

    def set_schema(self, schema: Dict[str, Any]) -> None:
        """Sets a custom schema for the ArangoDB Database."""
        self.__schema = schema

    def refresh_schema(
        self,
        sample_ratio: float = 0,
        graph_name: Optional[str] = None,
        include_examples: bool = True,
        list_limit: int = 32,
    ) -> None:
        """
        Refresh the graph schema information.

        Parameters:
        - sample_ratio (float): A float (0 to 1) to determine the
            ratio of documents/edges sampled in relation to the Collection size
            to generate each Collection Schema. If 0, one document/edge
            is used per Collection. Defaults to 0.
        - graph_name (str): The name of an existing ArangoDB Graph to specifically
            use to generate the schema. If None, the entire database will be used.
            Defaults to None.
        - include_examples (bool): Whether to include example values fetched from
            a sample documents as part of the schema. Defaults to True. Lists of size
            higher than **list_limit** will be excluded from the schema, even if
            **schema_include_examples** is set to True. Defaults to True.
        - list_limit (int): The maximum list size the schema will include as part
            of the example values. If the list is longer than this limit, a string
            describing the list will be used in the schema instead. Default is 32.
        """
        self.__schema = self.generate_schema(
            sample_ratio, graph_name, include_examples, list_limit
        )

    def generate_schema(
        self,
        sample_ratio: float = 0,
        graph_name: Optional[str] = None,
        include_examples: bool = True,
        list_limit: int = 32,
        schema_string_limit: int = 256,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        Generates the schema of the ArangoDB Database and returns it

        Parameters:
        - sample_ratio (float): A ratio (0 to 1) to determine the
        ratio of documents/edges used (in relation to the Collection size)
        to render each Collection Schema. If 0, one document/edge
        is used per Collection.
        - graph_name (str): The name of the graph to use to generate the schema. If
            None, the entire database will be used.
        - include_examples (bool): A flag whether to scan the database for
            example values and use them in the graph schema. Default is True.
        - list_limit (int): The maximum number of elements to include in a list.
            If the list is longer than this limit, a string describing the list
            will be used in the schema instead. Default is 32.
        - schema_string_limit (int): The maximum number of characters to include
            in a string. If the string is longer than this limit, a string
            describing the string will be used in the schema instead. Default is 128.
        """
        if not 0 <= sample_ratio <= 1:
            raise ValueError("**sample_ratio** value must be in between 0 to 1")

        graph_schema: List[Dict[str, Any]] = []
        if graph_name:
            # Fetch a single graph
            graph: Graph = self.db.graph(graph_name)
            edge_definitions = graph.edge_definitions()

            graph_schema = [{"name": graph_name, "edge_definitions": edge_definitions}]

            # Fetch graph-specific collections
            collection_names = set(graph.vertex_collections())  # type: ignore
            for edge_definition in edge_definitions:  # type: ignore
                collection_names.add(edge_definition["edge_collection"])

        else:
            # Fetch all graphs
            graph_schema = [
                {"graph_name": g["name"], "edge_definitions": g["edge_definitions"]}
                for g in self.db.graphs()  # type: ignore
            ]

            # Fetch all collections
            collection_names = {
                collection["name"]
                for collection in self.db.collections()  # type: ignore
            }

        # Stores the schema of every ArangoDB Document/Edge collection
        collection_schema: List[Dict[str, Any]] = []
        for collection in self.db.collections():  # type: ignore
            if collection["system"] or collection["name"] not in collection_names:
                continue

            # Extract collection name, type, and size
            col_name: str = collection["name"]
            col_type: str = collection["type"]
            col_size: int = self.db.collection(col_name).count()  # type: ignore

            # Set number of ArangoDB documents/edges to retrieve
            limit_amount = ceil(sample_ratio * col_size) or 1

            aql = f"""
                FOR doc in @@col_name
                    LIMIT {limit_amount}
                    RETURN doc
            """

            doc: Dict[str, Any]
            properties: List[Dict[str, str]] = []
            cursor = self.db.aql.execute(aql, bind_vars={"@col_name": col_name})
            for doc in cursor:  # type: ignore
                for key, value in doc.items():
                    properties.append({key: type(value).__name__})

            collection_schema_entry = {
                "name": col_name,
                "type": col_type,
                "size": col_size,
                "properties": properties,
            }

            if include_examples and col_size > 0:
                collection_schema_entry["example"] = self._sanitize_input(
                    doc, list_limit, schema_string_limit
                )

            collection_schema.append(collection_schema_entry)

        return {"graph_schema": graph_schema, "collection_schema": collection_schema}

    def query(
        self,
        query: str,
        params: dict = {},
    ) -> List[Any]:
        """
        Execute an AQL query and return the results.

        Parameters:
        - query (str): The AQL query to execute.
        - params (dict): Additional arguments piped to the function.
            - top_k: Number of results to process from the AQL cursor.
                Defaults to None.
            - list_limit: Removes lists above **list_limit** size
                that have been returned from the AQL query.
            - string_limit: Removes strings above **string_limit** size
                that have been returned from the AQL query.
            - Remaining params are passed to the AQL query execution.

        Returns:
        - A list of dictionaries containing the query results.
        """
        top_k = params.pop("top_k", None)
        list_limit = params.pop("list_limit", 32)
        string_limit = params.pop("string_limit", 256)
        cursor = self.__db.aql.execute(query, **params)

        results = []

        i = 0
        for doc in cursor:  # type: ignore
            results.append(self._sanitize_input(doc, list_limit, string_limit))

            if top_k and i >= top_k:
                break

            i += 1

        return results

    def explain(self, query: str, params: dict = {}) -> List[Dict[str, Any]]:
        """
        Explain an AQL query without executing it.

        Parameters:
        - query (str): The AQL query to explain.

        Returns:
        - A list of dictionaries containing the query explanation.
        """
        return self.__db.aql.explain(query)  # type: ignore

    def add_graph_documents(
        self,
        graph_documents: List[GraphDocument],
        include_source: bool = False,
        graph_name: Optional[str] = None,
        update_graph_definition_if_exists: bool = False,
        batch_size: int = 1000,
        use_one_entity_collection: bool = True,
        insert_async: bool = False,
        source_collection_name: Union[str, None] = None,
        source_edge_collection_name: Union[str, None] = None,
        entity_collection_name: Union[str, None] = None,
        entity_edge_collection_name: Union[str, None] = None,
        embeddings: Union[Embeddings, None] = None,
        embedding_field: str = "embedding",
        embed_source: bool = False,
        embed_nodes: bool = False,
        embed_relationships: bool = False,
        capitalization_strategy: str = "none",
    ) -> None:
        """
        Constructs nodes & relationships in the graph based on the
        provided GraphDocument objects.

        Parameters:
        - graph_documents (List[GraphDocument]): A list of GraphDocument objects
        that contain the nodes and relationships to be added to the graph. Each
        GraphDocument should encapsulate the structure of part of the graph,
        including nodes, relationships, and the source document information.
        - include_source (bool, optional): If True, stores the source document
        and links it to nodes in the graph using the HAS_SOURCE relationship.
        This is useful for tracing back the origin of data. Merges source
        documents based on the `id` property from the source document if available,
        otherwise it calculates the Farmhash hash of `page_content`
        for merging process. Defaults to False.
        - graph_name (str): The name of the ArangoDB General Graph to create. If None,
            no graph will be created.
        - update_graph_definition_if_exists (bool): If True, updates the graph
            Edge Definitions
        if it already exists. Defaults to False. Not used if `graph_name` is None. It is
        recommended to set this to True if `use_one_entity_collection` is set to False.
        - batch_size (int): The number of nodes/edges to insert in a single batch.
        - use_one_entity_collection (bool): If True, all nodes are stored in a single
        entity collection. If False, nodes are stored in separate collections based
        on their type. Defaults to True.
        - insert_async (bool): If True, inserts data asynchronously. Defaults to False.
        - source_collection_name (str): The name of the collection to store the source
        documents. Defaults to "SOURCE".
        - source_edge_collection_name (str): The name of the edge collection to store
        the relationships between source documents and nodes. Defaults to "HAS_SOURCE".
        - entity_collection_name (str): The name of the collection to store the nodes.
        Defaults to "ENTITY". Only used if `use_one_entity_collection` is True.
        - entity_edge_collection_name (str): The name of the edge collection to store
        the relationships between nodes. Defaults to "LINKS_TO". Only used if
        `use_one_entity_collection` is True.
        - embeddings (Embeddings): An Embeddings object to use for embedding the source,
        nodes and relationships. Defaults to None.
        - embedding_field (set[str]): The field name to store the embedding. Defaults
            to "embedding". Only used if `embedding` is not None, and `embed_source`,
            `embed_nodes`, or `embed_relationships` is True.
        - embed_source (bool): If True, embeds the source document. Defaults to False.
        - embed_nodes (bool): If True, embeds the nodes. Defaults to False.
        - embed_relationships (bool): If True, embeds the relationships.
            Defaults to False.
        - capitalization_strategy (str): The capitalization strategy applied on the
            node and edge keys. Can be "lower", "upper", or "none". Defaults to "none".
            Useful as a basic Entity Resolution technique to avoid duplicates based
            on capitalization.
        """
        if not graph_documents:
            return

        if not embeddings and (embed_source or embed_nodes or embed_relationships):
            m = "**embedding** is required to embed source, nodes, or relationships."
            raise ValueError(m)

        def embed_text(text: str) -> list[float]:
            if not embeddings:
                raise ValueError("**embedding** is required to embed text.")

            res: Any = embeddings.embed_documents([text])[0]

            while type(res) is list:
                if type(res[0]) is float:
                    break

                res = res[0]

            return res

        if capitalization_strategy == "none":
            capitalization_fn = lambda x: x  # noqa: E731
        if capitalization_strategy == "lower":
            capitalization_fn = str.lower
        elif capitalization_strategy == "upper":
            capitalization_fn = str.upper
        else:
            m = "**capitalization_strategy** must be 'lower', 'upper', or 'none'."
            raise ValueError(m)

        #########
        # Setup #
        #########

        suffix = f"{graph_name}_" if graph_name else ""
        if not source_collection_name:
            source_collection_name = f"{suffix}SOURCE"
        if not source_edge_collection_name:
            source_edge_collection_name = f"{suffix}HAS_SOURCE"
        if not entity_collection_name:
            entity_collection_name = f"{suffix}ENTITY"
        if not entity_edge_collection_name:
            entity_edge_collection_name = f"{suffix}LINKS_TO"

        insertion_db = self.__async_db if insert_async else self.__db
        nodes: DefaultDict[str, list[dict[str, Any]]] = defaultdict(list)
        edges: DefaultDict[str, list[dict[str, Any]]] = defaultdict(list)
        edge_definitions_dict: DefaultDict[str, DefaultDict[str, Set[str]]] = (
            defaultdict(lambda: defaultdict(set))
        )

        if include_source:
            self._create_collection(source_collection_name)
            self._create_collection(source_edge_collection_name, is_edge=True)

            from_cols = {entity_collection_name} if use_one_entity_collection else set()

            edge_definitions_dict[source_edge_collection_name][
                "from_vertex_collections"
            ] = from_cols
            edge_definitions_dict[source_edge_collection_name][
                "to_vertex_collections"
            ] = {source_collection_name}

        if use_one_entity_collection:
            self._create_collection(entity_collection_name)
            self._create_collection(entity_edge_collection_name, is_edge=True)

            edge_definitions_dict[entity_edge_collection_name][
                "from_vertex_collections"
            ] = {entity_collection_name}
            edge_definitions_dict[entity_edge_collection_name][
                "to_vertex_collections"
            ] = {entity_collection_name}

        process_node_fn = (
            self._process_node_as_entity
            if use_one_entity_collection
            else self._process_node_as_type
        )

        process_edge_fn = (
            self._process_edge_as_entity
            if use_one_entity_collection
            else self._process_edge_as_type
        )

        #######################
        # Document Processing #
        #######################

        for document in graph_documents:
            source_id_hash = None

            # 1. Process Source Document
            if include_source:
                if not document.source:
                    raise ValueError("Source document is required.")

                source_embedding = (
                    embed_text(document.source.page_content) if embed_source else None
                )

                source_id_hash = self._process_source(
                    document.source,
                    source_collection_name,
                    source_embedding,
                    embedding_field,
                    insertion_db,
                )

            # 2. Process Nodes
            node_key_map = {}
            for i, node in enumerate(document.nodes, 1):
                node.id = capitalization_fn(str(node.id))
                node_key = self._hash(node.id)
                node_key_map[node.id] = node_key

                if embed_nodes:
                    node.properties[embedding_field] = embed_text(node.id)

                node_type = process_node_fn(
                    node_key, node, nodes, entity_collection_name
                )

                # 2.1 Link Source Document to Node
                if include_source:
                    edges[source_edge_collection_name].append(
                        {
                            "_from": f"{node_type}/{node_key}",
                            "_to": f"{source_collection_name}/{source_id_hash}",
                        }
                    )

                    if not use_one_entity_collection:
                        edge_definitions_dict[source_edge_collection_name][
                            "from_vertex_collections"
                        ].add(node_type)

                # 2.2 Batch Insert
                if i % batch_size == 0:
                    self._import_data(insertion_db, nodes, is_edge=False)
                    self._import_data(insertion_db, edges, is_edge=True)

            self._import_data(insertion_db, nodes, is_edge=False)
            self._import_data(insertion_db, edges, is_edge=True)

            # 3. Process Edges
            for i, edge in enumerate(document.relationships, 1):
                source, target = edge.source, edge.target
                source.id = capitalization_fn(str(source.id))
                target.id = capitalization_fn(str(target.id))

                edge_str = f"{source.id} {edge.type} {target.id}"

                if embed_relationships:
                    edge.properties[embedding_field] = embed_text(edge_str)

                if include_source:
                    edge.properties["source_id"] = source_id_hash

                source_key = self._get_node_key(
                    source,
                    nodes,
                    node_key_map,
                    entity_collection_name,
                    process_node_fn,
                )

                target_key = self._get_node_key(
                    target,
                    nodes,
                    node_key_map,
                    entity_collection_name,
                    process_node_fn,
                )

                edge_key = self._hash(edge_str)
                process_edge_fn(
                    edge,
                    edge_str,
                    edge_key,
                    source_key,
                    target_key,
                    edges,
                    entity_collection_name,
                    entity_edge_collection_name,
                    edge_definitions_dict,
                )

                # 3.1 Batch Insert
                if i % batch_size == 0:
                    self._import_data(insertion_db, edges, is_edge=True)
                    self._import_data(insertion_db, nodes, is_edge=False)

            self._import_data(insertion_db, edges, is_edge=True)
            self._import_data(insertion_db, nodes, is_edge=False)

        ##################
        # Graph Creation #
        ##################

        if graph_name:
            edge_definitions: list[dict[str, Union[str, list[str]]]] = [
                {
                    "edge_collection": k,
                    "from_vertex_collections": list(v["from_vertex_collections"]),
                    "to_vertex_collections": list(v["to_vertex_collections"]),
                }
                for k, v in edge_definitions_dict.items()
            ]

            if not self.db.has_graph(graph_name):
                self.db.create_graph(graph_name, edge_definitions)

            elif update_graph_definition_if_exists:
                graph = self.db.graph(graph_name)

                for e_d in edge_definitions:
                    e_c = str(e_d["edge_collection"])
                    if not graph.has_edge_definition(e_c):
                        graph.create_edge_definition(**e_d)  # type: ignore
                    else:
                        graph.replace_edge_definition(**e_d)  # type: ignore

        # Refresh schema after insertions
        self.refresh_schema()

    @classmethod
    def from_db_credentials(
        cls,
        url: Optional[str] = None,
        dbname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Any:
        """Convenience constructor that builds Arango DB from credentials.

        Args:
            url: Arango DB url. Can be passed in as named arg or set as environment
                var ``ARANGODB_URL``. Defaults to "http://localhost:8529".
            dbname: Arango DB name. Can be passed in as named arg or set as
                environment var ``ARANGODB_DBNAME``. Defaults to "_system".
            username: Can be passed in as named arg or set as environment var
                ``ARANGODB_USERNAME``. Defaults to "root".
            password: Can be passed ni as named arg or set as environment var
                ``ARANGODB_PASSWORD``. Defaults to "".

        Returns:
            An arango.database.StandardDatabase.
        """
        db = get_arangodb_client(
            url=url, dbname=dbname, username=username, password=password
        )
        return cls(db)

    def _import_data(
        self,
        db: Database,
        data: Dict[str, List[Dict[str, Any]]],
        is_edge: bool,
    ) -> None:
        """Imports data into the ArangoDB database in bulk."""
        for collection, batch in data.items():
            self._create_collection(collection, is_edge)
            db.collection(collection).import_bulk(batch, on_duplicate="update")

        data.clear()

    def _create_collection(
        self, collection_name: str, is_edge: bool = False, **kwargs: Any
    ) -> None:
        """Creates a collection in the ArangoDB database if it does not exist."""
        if not self.db.has_collection(collection_name):
            self.db.create_collection(collection_name, edge=is_edge, **kwargs)

    def _process_node_as_entity(
        self,
        node_key: str,
        node: Node,
        nodes: DefaultDict[str, list],
        entity_collection_name: str,
    ) -> str:
        """Processes a Graph Document Node into ArangoDB as a unanimous Entity."""
        nodes[entity_collection_name].append(
            {
                "_key": node_key,
                "text": node.id,
                "type": node.type,
                **node.properties,
            }
        )
        return entity_collection_name

    def _process_node_as_type(
        self, node_key: str, node: Node, nodes: DefaultDict[str, list], _: str
    ) -> str:
        """Processes a Graph Document Node into ArangoDB based on its Node Type."""
        node_type = self._sanitize_collection_name(node.type)
        nodes[node_type].append({"_key": node_key, "text": node.id, **node.properties})
        return node_type

    def _process_edge_as_entity(
        self,
        edge: Relationship,
        edge_str: str,
        edge_key: str,
        source_key: str,
        target_key: str,
        edges: DefaultDict[str, list],
        entity_collection_name: str,
        entity_edge_collection_name: str,
        _: DefaultDict[str, DefaultDict[str, set[str]]],
    ) -> None:
        """Processes a Graph Document Edge into ArangoDB as a unanimous Entity."""
        edges[entity_edge_collection_name].append(
            {
                "_key": edge_key,
                "_from": f"{entity_collection_name}/{source_key}",
                "_to": f"{entity_collection_name}/{target_key}",
                "type": edge.type,
                "text": edge_str,
                **edge.properties,
            }
        )

    def _process_edge_as_type(
        self,
        edge: Relationship,
        edge_str: str,
        edge_key: str,
        source_key: str,
        target_key: str,
        edges: DefaultDict[str, list],
        _1: str,
        _2: str,
        edge_definitions_dict: DefaultDict[str, DefaultDict[str, set[str]]],
    ) -> None:
        """Processes a Graph Document Edge into ArangoDB based on its Edge Type."""
        source: Node = edge.source
        target: Node = edge.target

        edge_type = self._sanitize_collection_name(edge.type)
        source_type = self._sanitize_collection_name(source.type)
        target_type = self._sanitize_collection_name(target.type)

        edge_definitions_dict[edge_type]["from_vertex_collections"].add(source_type)
        edge_definitions_dict[edge_type]["to_vertex_collections"].add(target_type)

        edges[edge_type].append(
            {
                "_key": edge_key,
                "_from": f"{source_type}/{source_key}",
                "_to": f"{target_type}/{target_key}",
                "text": edge_str,
                **edge.properties,
            }
        )

    def _get_node_key(
        self,
        node: Node,
        nodes: DefaultDict[str, list],
        node_key_map: Dict[str, str],
        entity_collection_name: str,
        process_node_fn: Any,
    ) -> str:
        """Gets the key of a node and processes it if it doesn't exist."""
        node.id = str(node.id)
        if node.id in node_key_map:
            return node_key_map[node.id]

        node_key = self._hash(node.id)
        node_key_map[node.id] = node_key
        process_node_fn(node_key, node, nodes, entity_collection_name)

        return node_key

    def _process_source(
        self,
        source: Document,
        source_collection_name: str,
        source_embedding: Union[list[float], None],
        embedding_field: str,
        insertion_db: Database,
    ) -> str:
        """Processes a Graph Document Source into ArangoDB."""
        source_id = self._hash(
            source.id if source.id else source.page_content.encode("utf-8")
        )

        doc = {
            **source.metadata,
            "_key": source_id,
            "text": source.page_content,
            "type": source.type,
        }

        if source_embedding:
            doc[embedding_field] = source_embedding

        insertion_db.collection(source_collection_name).insert(doc, overwrite=True)

        return source_id

    def _hash(self, value: Any) -> str:
        """Applies the Farmhash hash function to a value."""
        try:
            value_str = str(value)
        except Exception:
            raise ValueError("Value must be a string or have a string representation.")

        return str(farmhash.Fingerprint64(value_str))  # type: ignore

    def _sanitize_collection_name(self, name: str) -> str:
        """
        Modifies a string to adhere to ArangoDB collection name rules.

        - Trims the name to 256 characters if it's too long.
        - Replaces invalid characters with underscores (_).
        - Ensures the name starts with a letter (prepends 'a' if needed).
        """
        if not name:
            raise ValueError("Collection name cannot be empty.")

        name = name[:256]

        # Replace invalid characters with underscores
        name = re.sub(r"[^a-zA-Z0-9_-]", "_", name)

        # Ensure the name starts with a letter; prepend 'a' if not
        if not re.match(r"^[a-zA-Z]", name):
            name = f"Collection_{name}"

        return name

    def _sanitize_input(self, d: Any, list_limit: int, string_limit: int) -> Any:
        """Sanitize the input dictionary or list.

        Sanitizes the input by removing embedding-like values,
        lists with more than **list_limit** elements, that are mostly irrelevant for
        generating answers in a LLM context. These properties, if left in
        results, can occupy significant context space and detract from
        the LLM's performance by introducing unnecessary noise and cost.

        Args:
            d (Any): The input dictionary or list to sanitize.
            list_limit (int): The limit for the number of elements in a list.
            string_limit (int): The limit for the number of characters in a string.

        Returns:
            Any: The sanitized dictionary or list.
        """

        if isinstance(d, dict):
            new_dict = {}
            for key, value in d.items():
                if isinstance(value, dict):
                    sanitized_value = self._sanitize_input(
                        value, list_limit, string_limit
                    )
                    if sanitized_value is not None:
                        new_dict[key] = sanitized_value
                elif isinstance(value, list):
                    if len(value) < list_limit:
                        sanitized_value = self._sanitize_input(
                            value, list_limit, string_limit
                        )
                        if sanitized_value is not None:
                            new_dict[key] = sanitized_value
                    else:
                        new_dict[key] = (
                            f"List of {len(value)} elements of type {type(value[0])}"
                        )
                elif isinstance(value, str):
                    if len(value) > string_limit:
                        new_dict[key] = f"String of {len(value)} characters"
                    else:
                        new_dict[key] = value
                else:
                    new_dict[key] = value
            return new_dict
        elif isinstance(d, list):
            if len(d) == 0:
                return d
            elif len(d) < list_limit:
                arr = []
                for item in d:
                    sanitized_item = self._sanitize_input(
                        item, list_limit, string_limit
                    )
                    if sanitized_item is not None:
                        arr.append(sanitized_item)
                return arr
            else:
                return f"List of {len(d)} elements of type {type(d[0])}"
        else:
            return d



from typing import Any, List, Union

from arango.database import StandardDatabase
from langchain_core.chat_history import BaseChatMessageHistory
from langchain_core.messages import BaseMessage, messages_from_dict


class ArangoChatMessageHistory(BaseChatMessageHistory):
    """Chat message history stored in an ArangoDB database."""

    def __init__(
        self,
        session_id: Union[str, int],
        db: StandardDatabase,
        collection_name: str = "ChatHistory",
        window: int = 3,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        # Make sure session id is not null
        if not session_id:
            raise ValueError("Please ensure that the session_id parameter is provided")

        self._session_id = str(session_id)
        self._db = db
        self._collection_name = collection_name
        self._window = window  # TODO: Use this

        if not self._db.has_collection(collection_name):
            self._db.create_collection(collection_name)

        self._collection = self._db.collection(self._collection_name)

        has_index = False
        for index in self._collection.indexes():  # type: ignore
            if "session_id" in index["fields"]:
                has_index = True
                break

        if not has_index:
            self._collection.add_persistent_index(["session_id"], unique=False)

        super().__init__(*args, **kwargs)

    @property
    def messages(self) -> List[BaseMessage]:
        """Retrieve the messages from ArangoDB"""
        query = """
            FOR doc IN @@col
                FILTER doc.session_id == @session_id
                SORT doc.time DESC
                RETURN UNSET(doc, ["session_id", "_id", "_rev"])
        """

        bind_vars = {"@col": self._collection_name, "session_id": self._session_id}

        cursor = self._db.aql.execute(query, bind_vars=bind_vars)  # type: ignore

        messages = [
            {"data": {"content": res["content"]}, "type": res["role"]}
            for res in cursor  # type: ignore
        ]

        return messages_from_dict(messages)

    @messages.setter
    def messages(self, messages: List[BaseMessage]) -> None:
        raise NotImplementedError(
            "Direct assignment to 'messages' is not allowed."
            " Use the 'add_messages' instead."
        )

    def add_message(self, message: BaseMessage) -> None:
        """Append the message to the record in ArangoDB"""

        self._db.collection(self._collection_name).insert(
            {
                "role": message.type,
                "content": message.content,
                "session_id": self._session_id,
            },
        )

    def clear(self) -> None:
        """Clear session memory from ArangoDB"""
        query = """
            FOR doc IN @@col
                FILTER doc.session_id == @session_id
                REMOVE doc IN @@col
        """

        bind_vars = {"@col": self._collection_name, "session_id": self._session_id}

        self._db.aql.execute(query, bind_vars=bind_vars)  # type: ignore
